==> app-0.7.5-20240920.151420-11.jar: monkey/ci/extensions.clj <==
(ns monkey.ci.extensions
  "Functionality for working with script extensions.  Extensions are a way
   for third party libraries to add functionality to scripts, that is easy
   to activate and can also be used in yaml-type scripts.  You could of 
   course also add regular functions to invoke, but this is not easy to
   use, especially when using container jobs.  Extensions do this by registering
   themselves under a specific namespaced keyword.  If this key is found in
   job properties, the associated extension code is executed.  Extensions can
   be executed before or after a job (or both)."
  (:require [clojure.tools.logging :as log]
            [manifold.deferred :as md]
            [monkey.ci
             [build :as b]
             [credits :as cr]
             [jobs :as j]
             [runtime :as rt]]))

(def new-register {})

(defonce registered-extensions (atom new-register))

(defn register [l ext]
  (assoc l (:key ext) ext))

(defn register! [ext]
  (swap! registered-extensions register ext))

(defmulti before-job (fn [k _] k))

(defmulti after-job (fn [k _] k))

(defn- find-mm
  "Finds multimethod for `mm` and given key, returns a fn that invokes it with runtime arg."
  [k mm]
  (when-let [m (get-method mm k)]
    #(m k %)))

(defn- run-safe [f ctx]
  (try
    (f ctx)
    (catch Exception ex
      (log/error "Failed to apply extension" ex)
      ;; Return context unchanged
      ctx)))

(defn- apply-extensions [{:keys [job] :as rt} registered rk mm]
  (->> (keys job)
       (reduce (fn [r k]
                 (let [b (or (get-in registered [k rk])
                             (find-mm k mm))]
                   (cond->> r
                     b (run-safe b))))
               rt)))

(defn apply-extensions-before
  ([rt registered]
   (apply-extensions rt registered :before before-job))
  ([rt]
   (apply-extensions-before rt @registered-extensions)))

(defn apply-extensions-after
  ([rt registered]
   (apply-extensions rt registered :after after-job))
  ([rt]
   (apply-extensions-after rt @registered-extensions)))

(defrecord ExtensionWrappingJob [target registered-ext]
  j/Job
  (execute! [job rt]
    (let [rt (assoc rt :job target)
          post-update (fn [rt]
                        (rt/post-events rt {:type :job/updated
                                            :sid (b/get-sid rt)
                                            :job (j/job->event (:job rt))})
                        rt)]
      ;; FIXME This is fairly dirty: jobs don't return the runtime, but extensions use it,
      ;; so maybe we need to think about reworking this.
      (-> rt
          (apply-extensions-before registered-ext)
          (as-> r (j/execute! target r))
          (md/chain
           ;; Add the result to the job in runtime
           (partial assoc-in rt [:job :result])
           ;; Dispatch an update event already.  Some extensions may require information
           ;; from the api that has not been sent yet, so we do it here.
           post-update
           ;; Let any extensions work on it
           #(apply-extensions-after % registered-ext)
           ;; Return the job result (possibly modified by extensions)
           (comp :result :job)))))

  cr/CreditConsumer
  (credit-multiplier [_ rt]
    (cr/calc-credit-multiplier target rt)))

(defn wrap-job
  "Wraps job so that extensions are invoked before and after it."
  ([job registered-ext]
   (map->ExtensionWrappingJob (-> job
                                  (j/job->event)
                                  (assoc :target job :registered-ext registered-ext))))
  ([job]
   (wrap-job job @registered-extensions)))

;;; Utility functions

(defn get-config
  "Retrieves configuration for the extension from the job"
  [rt k]
  (get-in rt [:job k]))

(defn set-value
  "Sets the extension value in the job.  This value will be stored with the job, so it
   must be serializable to `edn`."
  [rt k v]
  (assoc-in rt [:job :result k] v))

==> app-0.7.5-20240930.124654-30.jar: monkey/ci/extensions.clj <==
(ns monkey.ci.extensions
  "Functionality for working with script extensions.  Extensions are a way
   for third party libraries to add functionality to scripts, that is easy
   to activate and can also be used in yaml-type scripts.  You could of 
   course also add regular functions to invoke, but this is not easy to
   use, especially when using container jobs.  Extensions do this by registering
   themselves under a specific namespaced keyword.  If this key is found in
   job properties, the associated extension code is executed.  Extensions can
   be executed before or after a job (or both)."
  (:require [clojure.tools.logging :as log]
            [manifold.deferred :as md]
            [monkey.ci
             [build :as b]
             [credits :as cr]
             [jobs :as j]
             [runtime :as rt]]))

(def new-register {})

(defonce registered-extensions (atom new-register))

(defn register [l ext]
  (assoc l (:key ext) ext))

(defn register! [ext]
  (swap! registered-extensions register ext))

(defmulti before-job (fn [k _] k))

(defmulti after-job (fn [k _] k))

(defn- find-mm
  "Finds multimethod for `mm` and given key, returns a fn that invokes it with runtime arg."
  [k mm]
  (when-let [m (get-method mm k)]
    #(m k %)))

(defn- run-safe [f ctx]
  (try
    (f ctx)
    (catch Exception ex
      (log/error "Failed to apply extension" ex)
      ;; Return context unchanged
      ctx)))

(defn- apply-extensions [{:keys [job] :as rt} registered rk mm]
  (->> (keys job)
       (reduce (fn [r k]
                 (let [b (or (get-in registered [k rk])
                             (find-mm k mm))]
                   (cond->> r
                     b (run-safe b))))
               rt)))

(defn apply-extensions-before
  ([rt registered]
   (apply-extensions rt registered :before before-job))
  ([rt]
   (apply-extensions-before rt @registered-extensions)))

(defn apply-extensions-after
  ([rt registered]
   (apply-extensions rt registered :after after-job))
  ([rt]
   (apply-extensions-after rt @registered-extensions)))

(defrecord ExtensionWrappingJob [target registered-ext]
  j/Job
  (execute! [job rt]
    (let [rt (assoc rt :job target)]
      ;; FIXME This is fairly dirty: jobs don't return the runtime, but extensions use it,
      ;; so maybe we need to think about reworking this.
      (-> rt
          (apply-extensions-before registered-ext)
          (as-> r (j/execute! target r))
          (md/chain
           ;; Add the result to the job in runtime
           (partial assoc-in rt [:job :result])
           ;; Let any extensions work on it
           #(apply-extensions-after % registered-ext)
           ;; FIXME job/end event has already been sent at this point.  We need some way to get the extension
           ;; results in there.
           ;; Return the job result (possibly modified by extensions)
           (comp :result :job)))))

  cr/CreditConsumer
  (credit-multiplier [_ rt]
    (cr/calc-credit-multiplier target rt)))

(defn wrap-job
  "Wraps job so that extensions are invoked before and after it."
  ([job registered-ext]
   (map->ExtensionWrappingJob (-> job
                                  (j/job->event)
                                  (assoc :target job :registered-ext registered-ext))))
  ([job]
   (wrap-job job @registered-extensions)))

;;; Utility functions

(defn get-config
  "Retrieves configuration for the extension from the job"
  [rt k]
  (get-in rt [:job k]))

(defn set-value
  "Sets the extension value in the job.  This value will be stored with the job, so it
   must be serializable to `edn`."
  [rt k v]
  (assoc-in rt [:job :result k] v))
